@@ -3,26 +3,42 @@ from gzip import GzipFile
 from io import BytesIO
 from json import loads as json_loads
 from os import fsync
-from sys import exc_info, version
-from .utils import NoNumpyException  # keep 'unused' imports
-from .comment import strip_comment_line_with_symbol, strip_comments  # keep 'unused' imports
-from .encoders import TricksEncoder, json_date_time_encode, class_instance_encode, ClassInstanceEncoder, \
-	json_complex_encode, json_set_encode, numeric_types_encode, numpy_encode, nonumpy_encode, NoNumpyEncoder, \
-	nopandas_encode, pandas_encode  # keep 'unused' imports
-from .decoders import DuplicateJsonKeyException, TricksPairHook, json_date_time_hook, ClassInstanceHook, \
-	json_complex_hook, json_set_hook, numeric_types_hook, json_numpy_obj_hook, json_nonumpy_obj_hook, \
-	nopandas_hook, pandas_hook  # keep 'unused' imports
-from json import JSONEncoder
+from sys import exc_info
+
+from json_tricks.utils import is_py3
+from .utils import str_type, NoNumpyException  # keep 'unused' imports
+from .comment import strip_comments  # keep 'unused' imports
+from .encoders import TricksEncoder, json_date_time_encode, \
+	class_instance_encode, json_complex_encode, json_set_encode, numeric_types_encode, numpy_encode, \
+	nonumpy_encode, nopandas_encode, pandas_encode, noenum_instance_encode, \
+	enum_instance_encode, pathlib_encode # keep 'unused' imports
+from .decoders import TricksPairHook, \
+	json_date_time_hook, ClassInstanceHook, \
+	json_complex_hook, json_set_hook, numeric_types_hook, json_numpy_obj_hook, \
+	json_nonumpy_obj_hook, \
+	nopandas_hook, pandas_hook, EnumInstanceHook, \
+	noenum_hook, pathlib_hook, nopathlib_hook  # keep 'unused' imports
 
 
-is_py3 = (version[:2] == '3.')
-str_type = str if is_py3 else (basestring, unicode,)
 ENCODING = 'UTF-8'
 
 
 _cih_instance = ClassInstanceHook()
-DEFAULT_ENCODERS = [json_date_time_encode, json_complex_encode, json_set_encode, numeric_types_encode, class_instance_encode,]
-DEFAULT_HOOKS = [json_date_time_hook, json_complex_hook, json_set_hook, numeric_types_hook, _cih_instance,]
+_eih_instance = EnumInstanceHook()
+DEFAULT_ENCODERS = [json_date_time_encode, json_complex_encode, json_set_encode,
+					numeric_types_encode, class_instance_encode, ]
+DEFAULT_HOOKS = [json_date_time_hook, json_complex_hook, json_set_hook,
+				numeric_types_hook, _cih_instance, ]
+
+
+try:
+	import enum
+except ImportError:
+	DEFAULT_ENCODERS = [noenum_instance_encode,] + DEFAULT_ENCODERS
+	DEFAULT_HOOKS = [noenum_hook,] + DEFAULT_HOOKS
+else:
+	DEFAULT_ENCODERS = [enum_instance_encode,] + DEFAULT_ENCODERS
+	DEFAULT_HOOKS = [_eih_instance,] + DEFAULT_HOOKS
 
 try:
 	import numpy
@@ -43,13 +59,24 @@ else:
 	DEFAULT_ENCODERS = [pandas_encode,] + DEFAULT_ENCODERS
 	DEFAULT_HOOKS = [pandas_hook,] + DEFAULT_HOOKS
 
+try:
+    import pathlib
+except:
+    # No need to include a "nopathlib_encode" hook since we would not encounter
+    # the Path object if pathlib isn't available. However, we *could* encounter
+    # a serialized Path object (produced by a version of Python with pathlib).
+	DEFAULT_HOOKS = [nopathlib_hook,] + DEFAULT_HOOKS
+else:
+	DEFAULT_ENCODERS = [pathlib_encode,] + DEFAULT_ENCODERS
+	DEFAULT_HOOKS = [pathlib_hook,] + DEFAULT_HOOKS
 
-DEFAULT_NONP_ENCODERS = [nonumpy_encode,] + DEFAULT_ENCODERS    # DEPRECATED
-DEFAULT_NONP_HOOKS = [json_nonumpy_obj_hook,] + DEFAULT_HOOKS   # DEPRECATED
+
+DEFAULT_NONP_ENCODERS = [nonumpy_encode,] + DEFAULT_ENCODERS		# DEPRECATED
+DEFAULT_NONP_HOOKS = [json_nonumpy_obj_hook,] + DEFAULT_HOOKS		# DEPRECATED
 
 
 def dumps(obj, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_ENCODERS, extra_obj_encoders=(),
-		primitives=False, compression=None, allow_nan=False, conv_str_byte=False, **jsonkwargs):
+		primitives=False, compression=None, allow_nan=False, conv_str_byte=False, fallback_encoders=(), **jsonkwargs):
 	"""
 	Convert a nested data structure to a json string.
 
@@ -58,6 +85,7 @@ def dumps(obj, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_ENCODERS,
 	:param cls: The json encoder class to use, defaults to NoNumpyEncoder which gives a warning for numpy arrays.
 	:param obj_encoders: Iterable of encoders to use to convert arbitrary objects into json-able promitives.
 	:param extra_obj_encoders: Like `obj_encoders` but on top of them: use this to add encoders without replacing defaults. Since v3.5 these happen before default encoders.
+	:param fallback_encoders: These are extra `obj_encoders` that 1) are ran after all others and 2) only run if the object hasn't yet been changed.
 	:param allow_nan: Allow NaN and Infinity values, which is a (useful) violation of the JSON standard (default False).
 	:param conv_str_byte: Try to automatically convert between strings and bytes (assuming utf-8) (default False).
 	:return: The string containing the json-encoded version of obj.
@@ -68,7 +96,7 @@ def dumps(obj, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_ENCODERS,
 		raise TypeError('`extra_obj_encoders` should be a tuple in `json_tricks.dump(s)`')
 	encoders = tuple(extra_obj_encoders) + tuple(obj_encoders)
 	txt = cls(sort_keys=sort_keys, obj_encoders=encoders, allow_nan=allow_nan,
-		primitives=primitives, **jsonkwargs).encode(obj)
+		primitives=primitives, fallback_encoders=fallback_encoders, **jsonkwargs).encode(obj)
 	if not is_py3 and isinstance(txt, str):
 		txt = unicode(txt, ENCODING)
 	if not compression:
@@ -84,20 +112,25 @@ def dumps(obj, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_ENCODERS,
 
 
 def dump(obj, fp, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_ENCODERS, extra_obj_encoders=(),
-		 primitives=False, compression=None, force_flush=False, allow_nan=False, conv_str_byte=False, **jsonkwargs):
+		primitives=False, compression=None, force_flush=False, allow_nan=False, conv_str_byte=False,
+		fallback_encoders=(), **jsonkwargs):
 	"""
 	Convert a nested data structure to a json string.
 
 	:param fp: File handle or path to write to.
 	:param compression: The gzip compression level, or None for no compression.
 	:param force_flush: If True, flush the file handle used, when possibly also in the operating system (default False).
-	
+
 	The other arguments are identical to `dumps`.
 	"""
 	txt = dumps(obj, sort_keys=sort_keys, cls=cls, obj_encoders=obj_encoders, extra_obj_encoders=extra_obj_encoders,
-		primitives=primitives, compression=compression, allow_nan=allow_nan, conv_str_byte=conv_str_byte, **jsonkwargs)
+		primitives=primitives, compression=compression, allow_nan=allow_nan, conv_str_byte=conv_str_byte,
+				fallback_encoders=fallback_encoders, **jsonkwargs)
 	if isinstance(fp, str_type):
-		fh = open(fp, 'wb+')
+		if compression:
+			fh = open(fp, 'wb+')
+		else:
+			fh = open(fp, 'w+')
 	else:
 		fh = fp
 		if conv_str_byte:
@@ -106,9 +139,9 @@ def dump(obj, fp, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_ENCODE
 			except TypeError:
 				pass
 				# if not isinstance(txt, str_type):
-				# 	# Cannot write bytes, so must be in text mode, but we didn't get a text
-				# 	if not compression:
-				# 		txt = txt.decode(ENCODING)
+				#		# Cannot write bytes, so must be in text mode, but we didn't get a text
+				#		if not compression:
+				#			txt = txt.decode(ENCODING)
 			else:
 				try:
 					fh.write(u'')
@@ -116,7 +149,7 @@ def dump(obj, fp, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_ENCODE
 					if isinstance(txt, str_type):
 						txt = txt.encode(ENCODING)
 	try:
-		if 'b' not in getattr(fh, 'mode', 'b?') and not isinstance(txt, str_type) and compression:
+		if compression and 'b' not in getattr(fh, 'mode', 'b?') and not isinstance(txt, str_type):
 			raise IOError('If compression is enabled, the file must be opened in binary mode.')
 		try:
 			fh.write(txt)
@@ -169,13 +202,17 @@ def loads(string, preserve_order=True, ignore_comments=True, decompression=None,
 		if conv_str_byte:
 			string = string.decode(ENCODING)
 		else:
-			raise TypeError(('Cannot automatically encode object of type "{0:}" in `json_tricks.load(s)` since '
-				'the encoding is not known. You should instead encode the bytes to a string and pass that '
-				'string to `load(s)`, for example bytevar.encode("utf-8") if utf-8 is the encoding.').format(type(string)))
+			raise TypeError(('The input was of non-string type "{0:}" in `json_tricks.load(s)`. '
+				'Bytes cannot be automatically decoding since the encoding is not known. Recommended '
+				'way is to instead encode the bytes to a string and pass that string to `load(s)`, '
+				'for example bytevar.encode("utf-8") if utf-8 is the encoding. Alternatively you can '
+				'force an attempt by passing conv_str_byte=True, but this may cause decoding issues.')
+					.format(type(string)))
 	if ignore_comments:
 		string = strip_comments(string)
 	obj_pairs_hooks = tuple(obj_pairs_hooks)
 	_cih_instance.cls_lookup_map = cls_lookup_map or {}
+	_eih_instance.cls_lookup_map = cls_lookup_map or {}
 	hooks = tuple(extra_obj_pairs_hooks) + obj_pairs_hooks
 	hook = TricksPairHook(ordered=preserve_order, obj_pairs_hooks=hooks, allow_duplicates=allow_duplicates)
 	return json_loads(string, object_pairs_hook=hook, **jsonkwargs)
@@ -192,7 +229,14 @@ def load(fp, preserve_order=True, ignore_comments=True, decompression=None, obj_
 	"""
 	try:
 		if isinstance(fp, str_type):
-			with open(fp, 'rb') as fh:
+			if decompression is not None:
+				open_binary = bool(decompression)
+			else:
+				with open(fp, 'rb') as fh:
+					# This attempts to detect gzip mode; gzip should always
+					# have this header, and text json can't have it.
+					open_binary = (fh.read(2) == b'\x1f\x8b')
+			with open(fp, 'rb' if open_binary else 'r') as fh:
 				string = fh.read()
 		else:
 			string = fp.read()
